/*
 *
 * Copyright (c) 2003 by FORCE Computers
 *
 * Note that this file is based on parts of OpenIPMI
 * written by Corey Minyard <minyard@mvista.com>
 * of MontaVista Software. Corey's code was helpful
 * and many thanks go to him. He gave the permission
 * to use this code in OpenHPI under BSD license.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *     Thomas Kanngieser <thomas.kanngieser@fci.com>
 */

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <unistd.h>
#include <assert.h>
#include <poll.h>
#include <sys/time.h>

#include "ipmi_con_lan.h"
#include "ipmi_utils.h"
#include "ipmi_addr.h"
#include "ipmi_cmd.h"


cIpmiConLan::cIpmiConLan( struct in_addr addr, int port,
                          tIpmiAuthType auth, tIpmiPrivilege priv,
                          char *user, char *passwd )
  : m_fd( -1 ), m_slave_addr( 0x20 ), m_is_active( false ),
    m_port( port ), m_auth( auth ), m_priv( priv ), m_auth_method( 0 ),
    m_ping_count( 0 ), m_last_seq( 0 ),
    m_outbound_seq_num( 0 ), m_session_id( 0 ), m_working_auth( eIpmiAuthTypeNone ),
    m_recv_msg_map( 0 ), m_inbound_seq_num( 0 ),
    m_inuse( false ), m_retries_left( 0 ), m_use_orig_addr( false )
{
  assert( strlen( user ) < dIpmiUsernameMax );
  assert( strlen( passwd ) < dIpmiPasswordMax );

  m_ip_addr.sin_family = AF_INET;
  m_ip_addr.sin_port = htons( port );
  m_ip_addr.sin_addr = addr;
  m_port = port;

  memset( m_username, 0, dIpmiUsernameMax );
  strcpy( m_username, user );

  memset( m_passwd, 0, dIpmiPasswordMax );
  strcpy( m_passwd, passwd );
}


cIpmiConLan::~cIpmiConLan()
{
  if ( IsOpen() )
       Close();

  if ( m_auth_method )
       delete m_auth_method;
}


int
cIpmiConLan::AuthGen( unsigned char *out,
                      uint8_t       *ses_id,
                      uint8_t       *seq,
                      unsigned char *data,
                      unsigned int  data_len )
{
  assert( m_auth == m_working_auth );
  assert( m_auth_method );
  
  int rv;
  cIpmiAuthSg l[] =
  {
    { ses_id, 4 },
    { data,   data_len },
    { seq,    4 },
    { NULL,   0 }
  };

  rv = m_auth_method->Gen( l, out );

  return rv;
}


int
cIpmiConLan::AuthCheck( uint8_t       *ses_id,
                        uint8_t       *seq,
                        unsigned char *data,
                        unsigned int  data_len,
                        unsigned char *code )
{
  assert( m_auth == m_working_auth );
  assert( m_auth_method );

  int rv;
  cIpmiAuthSg l[] =
  {
    { ses_id, 4  },
    { data,   data_len },
    { seq,    4 },
    { NULL,   0 }
  };

  rv = m_auth_method->Check( l, code );

  return rv;
}




int
cIpmiConLan::OpenLanFd()
{
  int                fd;
  struct sockaddr_in addr;
  int                curr_port;
  int                rv;

  fd = socket( PF_INET, SOCK_DGRAM, IPPROTO_UDP );

  if ( fd == -1 )
       return fd;

  curr_port = 7000;
  do
     {
       curr_port++;
       addr.sin_family = AF_INET;
       addr.sin_port = htons(curr_port);
       addr.sin_addr.s_addr = INADDR_ANY;

       rv = bind( fd, (struct sockaddr *)&addr, sizeof( addr ) );
     }
  while( curr_port < 7100 && rv == -1 );

  if ( rv == -1 )
     {
       int tmp_errno = errno;
       close( fd );
       errno = tmp_errno;
       return -1;
     }

  IpmiLog( "using port %d.\n", curr_port );

  return fd;
}


unsigned char
cIpmiConLan::Checksum( unsigned char *data, int size )
{
  unsigned char csum = 0;

  for( ; size > 0; size--, data++ )
       csum += *data;

  return -csum;
}


#define dIpmiMaxLanLen (dIpmiMaxMsgLength + 42)

int
cIpmiConLan::SendMsg( const cIpmiAddr &addr, const cIpmiMsg &msg )
{
  cIpmiAddr tmp_addr = addr;

  if ( m_inuse )
     {
       assert( !m_inuse );
       return EBUSY;
     }

  unsigned int seq = (m_last_seq + 1) % 64;

  m_use_orig_addr = false;

  if (    addr.m_type == eIpmiAddrTypeIpmb
       || addr.m_type == eIpmiAddrTypeIpmbBroadcast )
     {
       if ( addr.m_slave_addr == m_slave_addr )
	  {
	    // Most systems don't handle sending to your own slave
	    // address, so we have to translate here.

            tmp_addr.Si();
            tmp_addr.m_lun = addr.m_lun;

            m_use_orig_addr = true;
            m_orig_addr     = addr;

            // In case it's a broadcast.
            m_orig_addr.m_type = eIpmiAddrTypeIpmb;
	  }
     }

  m_last_seq = seq;

  m_inuse = true;
  m_addr  = tmp_addr;
  m_msg   = msg;

  if ( addr.m_type == eIpmiAddrTypeIpmbBroadcast )
       m_retries_left = 0;
  else
       m_retries_left = dLanRspRetries;

  int rv = Send( tmp_addr, msg );

  if ( rv )
       m_inuse = false;

  return rv;
}


int
cIpmiConLan::SendPing()
{
  unsigned char data[dIpmiMaxLanLen];

  data[0]  = 6; // RMCP version 1.0.
  data[1]  = 0;
  data[2]  = 0xff; // no RMCP ACK
  data[3]  = 0x06; // ASF
  IpmiSetUint32( data + 4,  dAsfIana );
  data[8]  = 0x80; // presence ping
  data[9]  = 0xff; // ????
  data[10] = 0x00;
  data[11] = 0x00;

  IpmiLog( "sending RMCP ping.\n" );

  int rv = sendto( m_fd, data, 12, 0,
                   (struct sockaddr *)&m_ip_addr,
                   sizeof( struct sockaddr_in ) );

  if ( rv == -1 )
       return errno;

  m_ping_count += 1;

  return 0;
}


int
cIpmiConLan::Send( const cIpmiAddr &addr, const cIpmiMsg &msg )
{
  unsigned char  data[dIpmiMaxLanLen];
  unsigned char *tmsg;
  int            pos;
  int            msgstart;
  int            rv;

  switch( addr.m_type ) 
     {
       case eIpmiAddrTypeSystemInterface:
       case eIpmiAddrTypeIpmb:
       case eIpmiAddrTypeIpmbBroadcast:
	    break;
       default:
            assert( 0 );
	    return EINVAL;
     }

  data[0] = 6; // RMCP version 1.0.
  data[1] = 0;
  data[2] = 0xff;
  data[3] = 0x07;
  data[4] = m_working_auth;

  IpmiSetUint32( data+5, m_outbound_seq_num );
  IpmiSetUint32( data+9, m_session_id );

  if ( m_working_auth == 0 )
       tmsg = data + 14;
  else
       tmsg = data + 30;

  if ( addr.m_type == eIpmiAddrTypeSystemInterface )
     {
       // It's a message straight to the BMC.
       tmsg[0] = 0x20; // To the BMC.
       tmsg[1] = (msg.m_netfn << 2) | addr.m_lun;
       tmsg[2] = Checksum( tmsg, 2 );
       tmsg[3] = 0x81; // Remote console IPMI Software ID
       tmsg[4] = m_last_seq << 2;
       tmsg[5] = msg.m_cmd;
       memcpy( tmsg + 6, msg.m_data, msg.m_data_len );
       pos = msg.m_data_len + 6;
       tmsg[pos] = Checksum( tmsg + 3, pos - 3 );
       pos++;
     }
  else
     {
       // It's an IPMB address, route it using a send message
       // command.
       pos = 0;
       tmsg[pos++] = 0x20; // BMC is the bridge.
       tmsg[pos++] = (eIpmiNetfnApp << 2) | 0;
       tmsg[pos++] = Checksum( tmsg, 2 );
       tmsg[pos++] = 0x81; // Remote console IPMI Software ID
       tmsg[pos++] = m_last_seq << 2;
       tmsg[pos++] = eIpmiCmdSendMsg;
       tmsg[pos++] = addr.m_channel & 0xf;

       if ( addr.m_type == eIpmiAddrTypeIpmbBroadcast )
	    tmsg[pos++] = 0; // Do a broadcast.

       msgstart = pos;
       tmsg[pos++] = addr.m_slave_addr;
       tmsg[pos++] = (msg.m_netfn << 2) | addr.m_lun;
       tmsg[pos++] = Checksum( tmsg + msgstart, 2 );
       msgstart = pos;
       tmsg[pos++] = 0x81;
       tmsg[pos++] = m_last_seq << 2;
       tmsg[pos++] = msg.m_cmd;
       memcpy( tmsg + pos, msg.m_data, msg.m_data_len );
       pos += msg.m_data_len;
       tmsg[pos] = Checksum( tmsg + msgstart, pos - msgstart );
       pos++;
       tmsg[pos] = Checksum( tmsg + 3, pos - 3 );
       pos++;
     }

  if ( m_working_auth == 0 )
     {
       // No authentication, so no authcode.
       data[13] = pos;
       pos += 14; // Convert to pos in data
     } 
  else
     {
       data[29] = pos;

       rv = AuthGen( data+13, data+9, data+5, tmsg, pos );

       if ( rv )
	    return rv;

       pos += 30; // Convert to pos in data
     }

  // Increment the outbound number, but make sure it's not zero.  If
  // it's already zero, ignore it, we are in pre-setup.
  if ( m_outbound_seq_num != 0 )
     {
       m_outbound_seq_num++;

       if ( m_outbound_seq_num == 0 )
	    m_outbound_seq_num++;
     }

  /*
    if (DEBUG_MSG) {
    ipmi_log(IPMI_LOG_DEBUG_START, "outgoing seq %d\n addr =", seq);
    dump_hex((unsigned char *) &(lan->ip_addr[lan->curr_ip_addr]),
    sizeof(struct sockaddr_in));
    ipmi_log(IPMI_LOG_DEBUG_CONT, "\n data =\n  ");
    dump_hex(data, pos);
    ipmi_log(IPMI_LOG_DEBUG_END, "");
    }
  */
  rv = sendto( m_fd, data, pos, 0,
	       (struct sockaddr *)&m_ip_addr,
	       sizeof(struct sockaddr_in) );

  if ( rv == -1 )
       rv = errno;
  else
       rv = 0;

  return rv;
}


cIpmiConLan::tResponseType
cIpmiConLan::HandleData( int fd,
                         cIpmiAddr &addr,
                         cIpmiMsg &msg )
{
  unsigned char       data[dIpmiMaxLanLen];
  struct sockaddr     ipaddrd;
  struct sockaddr_in *ipaddr;
  int                 len;
  socklen_t           from_len;
  uint32_t            seq, sess_id;
  unsigned char      *tmsg;
  unsigned int        data_len;

  from_len = sizeof( ipaddrd );
  len = recvfrom( fd, data, dIpmiMaxLanLen, 0, &ipaddrd, &from_len );

  if ( len < 0 )
       return eResponseOk;

  //IpmiLog( "rmcp msg: " );
  //IpmiLogHex( data, len );

  // Make sure the source IP matches what we expect the other end to
  // be.
  ipaddr = (struct sockaddr_in *)&ipaddrd;

  if (    (ipaddr->sin_port != m_ip_addr.sin_port)
       || (ipaddr->sin_addr.s_addr != m_ip_addr.sin_addr.s_addr) )
     {
       IpmiLog( "Dropped message due to invalid IP !\n");
       return eResponseOk;
     }

  // Validate the length first, so we know that all the data in the
  // buffer we will deal with is valid.
  if ( len < 21 )
     { 
       // Minimum size of an IPMI msg.
       IpmiLog( "Dropped message because too small(1)\n" );
       return eResponseOk;
     }

  // Validate the RMCP portion of the message.
  if (    data[0] != 6
       || data[2] != 0xff )
     {
       IpmiLog( "Dropped message not valid IPMI/RMCP !\n");
       return eResponseOk;
     }

  if ( data[3] == 0x06 )
     {
       unsigned int asf_iana = IpmiGetUint32( data+4 );
       
       if ( asf_iana != dAsfIana || data[8] != 0x40 )
          {
            IpmiLog( "Dropped message not valid RMCP pong message %04x, %04x, %02x !\n",
                     asf_iana, dAsfIana, data[8] );
            return eResponseOk;
          }

       m_ping_count--;

       IpmiLog( "reading RMCP pong.\n" );

       return eResponsePong;
     }

  if ( data[3] != 0x07 )
     {
       IpmiLog( "Dropped message not valid IPMI/RMCP\n");
       return eResponseOk;
     }
  
  if ( data[4] == 0 )
     {
       // No authentication.
       if ( len < data[13] + 14 ) 
	  {
	    // Not enough data was supplied, reject the message.
	    IpmiLog( "Dropped message because too small(2)\n");
	    return eResponseOk;
	  }

       data_len = data[13];
     }
  else
     {
       if ( len < 37 )
	  { 
	    // Minimum size of an authenticated IPMI msg.
	    IpmiLog( "Dropped message because too small(3)\n");
	    return eResponseOk;
	  }

       // authcode in message, add 16 to the above checks.
       if ( len < data[29] + 30 )
	  {
	    // Not enough data was supplied, reject the message.
	    IpmiLog( "Dropped message because too small(4)\n");
	    return eResponseOk;
	  }

       data_len = data[29];
     }

  // Drop if the authtypes are incompatible.
  if ( m_working_auth != data[4] )
     {
       IpmiLog( "Dropped message not valid authtype\n");
       return eResponseOk;
     }

  // Drop if sessions ID's don't match.
  sess_id = IpmiGetUint32( data+9 );
  if ( sess_id != m_session_id )
     {
       IpmiLog( "Dropped message not valid session id 0x%x != 0x%x\n",
		sess_id, m_session_id );
       return eResponseOk;
     }

  seq = IpmiGetUint32( data+5 );

  // dbg( "reading data: sid = 0x%x, seq = 0x%x", sess_id, seq );
 
  if ( data[4] != 0 )
     {
       // Validate the message's authcode.  Do this before checking
       // the session seq num so we know the data is valid.
       int rv = AuthCheck( data+9, data+5, data+30, data[29], data+13 );

       if (rv) 
          {
            IpmiLog( "Dropped message auth fail !\n");
            return eResponseOk;
          }

       tmsg = data + 30;
     }
  else
       tmsg = data + 14;

  // Check the sequence number.
  if ( seq - m_inbound_seq_num <= 8 )
     {
       // It's after the current sequence number, but within 8.  We
       // move the sequence number forward.
       m_recv_msg_map <<= seq - m_inbound_seq_num;
       m_recv_msg_map |= 1;
       m_inbound_seq_num = seq;
     }
  else if ( m_inbound_seq_num - seq <= 8 )
     {
       // It's before the current sequence number, but within 8.
       uint8_t bit = 1 << (m_inbound_seq_num - seq);
       if ( m_recv_msg_map & bit )
	  {
	    IpmiLog( "Dropped message duplicate\n" );
	    return eResponseOk;
	  }

       m_recv_msg_map |= bit;
     } 
  else
     {
       // It's outside the current sequence number range, discard
       // the packet.
       IpmiLog( "Dropped message out of seq range\n");
       return eResponseOk;
     }

  // Now we have an authentic in-sequence message.

  // We don't check the checksums, because the network layer should
  // validate all this for us.

  seq = tmsg[4] >> 2;

  if ( tmsg[5] == eIpmiCmdSendMsg )
     {
       addr = m_addr;

       // It's a response to a sent message.

       // FIXME - this entire thing is a cheap hack.
       if ( tmsg[6] != 0 )
	  {
	    // Got an error from the send message.  We don't have any
	    // IPMB information to work with, so just extract it from
            // the message.

            // Just in case it's a broadcast.
	    addr.m_type = eIpmiAddrTypeIpmb;

            msg.m_netfn    = (tIpmiNetfn)(m_msg.m_netfn | 1);
            msg.m_cmd      = m_msg.m_cmd;
            msg.m_data[0]  = tmsg[6];
            msg.m_data_len = 1;
	  }
       else
	  {
            addr.m_type = eIpmiAddrTypeIpmb;

            // This is a hack, but the channel does not come back in the
            // message.  So we use the channel from the original
	    // instead.
	    addr.m_slave_addr = tmsg[9];
	    addr.m_lun        = tmsg[10] & 0x3;

	    msg.m_netfn    = (tIpmiNetfn)(tmsg[7] >> 2);
	    msg.m_cmd      = (tIpmiCmd)tmsg[11];
	    msg.m_data_len = data_len - 14;
	    memcpy( msg.m_data, tmsg + 12, msg.m_data_len );
	  }
     }
  else if ( tmsg[5] == eIpmiCmdReadEventMsgBuffer )
     {
       cIpmiAddr ipmb_addr;

       if ( tmsg[6] != 0 )
	  {
	    // An error getting the events, just ignore it.
	    IpmiLog( "Dropped message err getting event\n");
	    return eResponseOk;
	  }

       ipmb_addr.m_type       = eIpmiAddrTypeIpmb;
       ipmb_addr.m_slave_addr = tmsg[3];
       ipmb_addr.m_lun        = tmsg[4] & 0x3;
       ipmb_addr.m_channel    = 0;

       msg.m_netfn    = (tIpmiNetfn)(tmsg[1] >> 2);
       msg.m_cmd      = (tIpmiCmd)tmsg[5];
       msg.m_data_len = data_len - 6 - 2; /* Remove completion code and checksum */
       memcpy( msg.m_data, tmsg + 6 + 1, msg.m_data_len );

       HandleEvent( ipmb_addr, msg );

       return eResponseEvent;
     }
  else
     {
       // It's a response directly from the BMC.
       addr.m_type    = eIpmiAddrTypeSystemInterface;
       addr.m_channel = 0xf;
       addr.m_lun     = tmsg[1] & 3;

       msg.m_netfn    = (tIpmiNetfn)(tmsg[1] >> 2);
       msg.m_cmd      = (tIpmiCmd)tmsg[5];
       msg.m_data_len = data_len - 6 - 1; /* Remove the checksum */
       memcpy( msg.m_data, tmsg+6, msg.m_data_len );
     }

  if ( !m_inuse )
     {
       IpmiLog( "Dropped message seq not in use\n");

       return eResponseError;
     }

  // Convert broadcast addresses to regular IPMB addresses, since
  // they come back that way.
  cIpmiAddr addr2 = m_addr;

  if ( addr2.m_type == eIpmiAddrTypeIpmbBroadcast )
       addr2.m_type = eIpmiAddrTypeIpmb;

  /* Validate that this response if for this command. */
  if (    (m_msg.m_netfn | 1) != msg.m_netfn
       || m_msg.m_cmd != msg.m_cmd
       || addr2 != addr )
     {
       IpmiLog( "Dropped message netfn/cmd/addr mismatch\n" );
       IpmiLog( "  netfn=%d, exp netfn=%d\n", msg.m_netfn, m_msg.m_netfn | 1 );
       IpmiLog( "  cmd=%d, exp cmd=%d\n", msg.m_cmd, m_msg.m_cmd );
       IpmiLog( "  addr    =" );
       addr.Log();
       IpmiLog( "\n" );
       IpmiLog( "  exp addr=" );
       addr2.Log();
       IpmiLog( "\n%d\n", len );
       IpmiLogHex( data, len );

       /*
	 dump_hex(addr, *addr_len);
	 dbg( "\n  exp addr=");
	 dump_hex( &addr2, m_seq_table[seq].addr_len);
	 dbg( "");
       */
       return eResponseOk;
     }

  if ( m_use_orig_addr )
       // We did an address translation, so translate back.
       addr = m_orig_addr;

  m_inuse = 0;

  return eResponseResponse;
}


cIpmiConLan::tResponseType
cIpmiConLan::WaitForResponse( unsigned int timeout_ms,
                              cIpmiAddr &addr, cIpmiMsg &msg )
{
  struct timeval tv;
  struct timeval timeout;
  struct timeval t0;
  int rv;
  tResponseType ret = eResponseOk;

  // create absolute timeout
  gettimeofday( &timeout, 0 );

  timeout.tv_sec += timeout_ms / 1000;
  timeout.tv_usec += (timeout_ms % 1000) * 1000;

  while( timeout.tv_usec > 10000000 )
     {
       timeout.tv_sec++;
       timeout.tv_usec -= 10000000;
     }

  // loop
  do
     {
       struct pollfd pfd;
                
       pfd.fd = m_fd;
       pfd.events = POLLIN;

       // relative timeout
       gettimeofday( &t0, 0 );

       timersub( &timeout, &t0, &tv );

       if ( tv.tv_sec < 0 || tv.tv_usec < 0 )
	  {
	    tv.tv_sec  = 0;
	    tv.tv_usec = 0;
	  }

       timeout_ms =  tv.tv_sec * 1000 + tv.tv_usec / 1000;

       rv = poll( &pfd, 1, timeout_ms );

       /* timeout */
       if ( !rv )
	    return eResponseTimeout;

       if ( rv == -1 )
	  {
	    assert( 0 );
	    return eResponseError;
	  }

       assert( rv == 1 );

       ret = (tResponseType)(ret | HandleData( m_fd, addr, msg ));

       // get currect time t0
       gettimeofday( &t0, 0 );
     }
  while( timercmp( &t0, &timeout, < ) < 0 && (ret & (eResponseResponse|eResponsePong)) == 0 );

  // timeout and no response ?
  if ( ret == 0 && timercmp( &t0, &timeout, < ) >= 0 )
       ret = (tResponseType)(ret | eResponseTimeout);

  return ret;
}


int
cIpmiConLan::SendMsgAndWaitForResponse( const cIpmiAddr &addr, const cIpmiMsg &msg,
                                        cIpmiAddr &rsp_addr, cIpmiMsg &rsp_msg )
{
  int i;

  for( i = 0; i < dLanRspRetries; i++ )
     {
       int rv = SendMsg( addr, msg );

       if ( rv )
          {
            m_inuse = 0;
            continue;
          }

       tResponseType res;

       do
          {
            res = WaitForResponse( dLanRspTimeout, rsp_addr,
                                   rsp_msg );
          }
       while( res == eResponseEvent );

       m_inuse = 0;

       if ( res & eResponseResponse )
            return 0;

       if ( res & eResponseError )
            continue;

       // handle pending events
       Event();

       // resend message
       IpmiLog( "resending RMCP msg.\n" );
     }

  return ETIMEDOUT;
}


int
cIpmiConLan::IfCmd( const cIpmiAddr &addr, const cIpmiMsg &msg,
                    cIpmiAddr &rsp_addr, cIpmiMsg &rsp_msg )
{
  while( 1 )
     {
       int rv = SendMsgAndWaitForResponse( addr, msg,
                                           rsp_addr, rsp_msg );

       if ( !rv )
            return 0;

       if ( rv != ETIMEDOUT )
            return rv;

       // check connection
       IpmiLog( "checking connection.\n" );

       cIpmiAddr  ping_addr;
       ping_addr.Si();
       cIpmiMsg   ping_msg( eIpmiNetfnApp, eIpmiCmdGetDeviceId );

       cIpmiAddr  ping_rsp_addr;
       cIpmiMsg   ping_rsp_msg;

       rv = SendMsgAndWaitForResponse( ping_addr, ping_msg, 
                                       ping_rsp_addr, ping_rsp_msg );

       if ( rv == 0 )
          {
            IpmiLog( "connection seems to be ok.\n" );
            return ETIMEDOUT;
          }

       // reconnect
       Reconnect();
     }

  return EINVAL;
}


int
cIpmiConLan::AuthCap()
{
  int       rv;
  cIpmiAddr addr( eIpmiAddrTypeSystemInterface );
  cIpmiMsg  msg( eIpmiNetfnApp, eIpmiCmdGetChannelAuthCapabilities );
  cIpmiAddr rsp_addr;
  cIpmiMsg  rsp_msg;

  msg.m_data[0] = 0xe;
  msg.m_data[1] = m_priv;
  msg.m_data_len = 2;

  // dbg( "RMCP: sending GetChannelAuthCapabilitiesCmd" );

  rv = SendMsgAndWaitForResponse( addr, msg,
                                  rsp_addr, rsp_msg );

  if ( rv )
       return rv;

  if (    (rsp_msg.m_data[0] != 0 )
       || (rsp_msg.m_data_len < 9 ) )
     {
       IpmiLog( "auth response = 0x%02x !\n", rsp_msg.m_data[0] );
       return EINVAL;
     }

  if ( !( rsp_msg.m_data[2] & (1 << m_auth ) ) )
     {
       IpmiLog( "Requested authentication not supported !\n" );

       char str[256] = "";

       if ( rsp_msg.m_data[2] & ( 1 << eIpmiAuthTypeNone ) )
            strcat( str, " none" );

       if ( rsp_msg.m_data[2] & ( 1 << eIpmiAuthTypeMd2 ) )
            strcat( str, " md2" );

       if ( rsp_msg.m_data[2] & ( 1 << eIpmiAuthTypeMd5 ) )
            strcat( str, " md5" );

       if ( rsp_msg.m_data[2] & ( 1 << eIpmiAuthTypeStraight ) )
            strcat( str, " straight" );

       if ( rsp_msg.m_data[2] & ( 1 << eIpmiAuthTypeOem ) )
            strcat( str, " oem" );

       IpmiLog( "supported authentication types:%s.\n", str );

       return EINVAL;
     }

  return 0;
}


int
cIpmiConLan::SetSessionPriv()
{
  int       rv;
  cIpmiAddr addr( eIpmiAddrTypeSystemInterface );
  cIpmiMsg  msg( eIpmiNetfnApp, eIpmiCmdSetSessionPrivilege );
  cIpmiAddr rsp_addr;
  cIpmiMsg  rsp_msg;

  msg.m_data[0]  = m_priv;
  msg.m_data_len = 1;

  // dbg( "RMCP: sending SetSessionPrivilegeCmd" );

  rv = SendMsgAndWaitForResponse( addr, msg,
                                  rsp_addr, rsp_msg );

  if ( rv )
       return rv;

  if ( rsp_msg.m_data[0] != 0 )
     {
       IpmiLog( "set session priv: 0x%02x !\n", rsp_msg.m_data[0] );
       return EINVAL;
     }

  if ( rsp_msg.m_data_len < 2 )
     {
       IpmiLog( "set session priv: msg to small: %d !\n", rsp_msg.m_data_len );
       return EINVAL;
     }

  if ( (unsigned char)m_priv != (rsp_msg.m_data[1] & 0xf))
     {
       // Requested privilege level did not match.
       IpmiLog( "set session priv: Requested privilege level did not match: %d, %d !\n",
                m_priv, (rsp_msg.m_data[1] & 0xf ) );

       return EINVAL;
     }

  return 0;
}


int
cIpmiConLan::ActiveSession()
{
  int       rv;
  cIpmiAddr addr( eIpmiAddrTypeSystemInterface );
  cIpmiMsg  msg( eIpmiNetfnApp, eIpmiCmdActivateSession );
  cIpmiAddr rsp_addr;
  cIpmiMsg  rsp_msg;

  msg.m_data[0] = m_auth;
  msg.m_data[1] = m_priv;
  memcpy( msg.m_data + 2, m_challenge_string, 16 );
  IpmiSetUint32( msg.m_data + 18, m_inbound_seq_num );
  msg.m_data_len = 22;

  // dbg( "RMCP: sending ActivateSessionCmd" );

  rv = SendMsgAndWaitForResponse( addr, msg,
                                  rsp_addr, rsp_msg );

  if ( rv )
       return rv;

  if ( rsp_msg.m_data[0] != 0 )
     {
       IpmiLog( "active session: 0x%02x !\n", rsp_msg.m_data[0] );
       return EINVAL;
     }

  if ( rsp_msg.m_data_len < 11 )
     {
       IpmiLog( "active session: msg to small: %d !\n", rsp_msg.m_data_len );
       return EINVAL;
     }

  m_working_auth = (tIpmiAuthType)(rsp_msg.m_data[1] & 0xf);

  if (    m_working_auth != 0
       && m_working_auth != m_auth )
     {
       // Eh?  It didn't return a valid authtype.
       IpmiLog( "active session: wrong auth: %d !\n", m_working_auth );
       return EINVAL;
     }

  m_session_id = IpmiGetUint32( rsp_msg.m_data + 2 );
  m_outbound_seq_num = IpmiGetUint32( rsp_msg.m_data + 6 );

  //        dbg( "reading: sid = 0x%x, seq = 0x%x", 
  //             m_session_id, m_outbound_seq_num );
        
  return 0;
}


int
cIpmiConLan::Challange()
{
  int       rv;
  cIpmiAddr addr( eIpmiAddrTypeSystemInterface );
  cIpmiMsg  msg( eIpmiNetfnApp, eIpmiCmdGetSessionChallenge );
  cIpmiAddr rsp_addr;
  cIpmiMsg  rsp_msg;

  msg.m_data[0]  = m_auth;
  msg.m_data_len = 1;
  memcpy( msg.m_data + 1, m_username, dIpmiUsernameMax );
  msg.m_data_len += dIpmiUsernameMax;

  // dbg( "RMCP: sending GetSessionChallengeCmd" );

  rv = SendMsgAndWaitForResponse( addr, msg,
                                  rsp_addr, rsp_msg );

  if ( rv )
       return rv;

  if ( rsp_msg.m_data[0] != 0 )
     {
       IpmiLog( "Challange returns: 0x%02x !\n",  rsp_msg.m_data[0] );
       return EINVAL;
     }

  if ( rsp_msg.m_data_len < 21 )
     {
       IpmiLog( "Challange response to small !\n" );
       return EINVAL;
     }

  // Get the temporary session id.
  m_session_id = IpmiGetUint32( rsp_msg.m_data + 1 );

  m_outbound_seq_num = 0;
  m_working_auth     = m_auth;
  memcpy( m_challenge_string, rsp_msg.m_data + 5, 16 );

  // Get a random number of the other end to start sending me sequence
  // numbers at, but don't let it be zero.
  while( m_inbound_seq_num == 0 )
       m_inbound_seq_num = random();

  return 0;
}


int
cIpmiConLan::IfOpen()
{
  int rv;

  m_ping_count       = 0;
  m_last_seq         = 0;
  m_outbound_seq_num = 0;
  m_session_id       = 0;
  m_working_auth     = eIpmiAuthTypeNone;
  m_recv_msg_map     = 0;
  m_inbound_seq_num  = 0;
  m_inuse            = 0;
  m_retries_left     = 0;
  m_use_orig_addr    = 0;

  m_auth_method = IpmiAuthFactory( m_auth );

  if ( m_auth_method == 0 )
     {
       IpmiLog( "unknown authentication method %d!\n", m_auth );
       return 0;
     }

  m_auth_method->Init( (unsigned char *)m_passwd );

  m_fd = OpenLanFd();

  if ( m_fd < 0 )
       return errno;

  rv = AuthCap();

  if ( rv )
       return rv;

  rv = Challange();
        
  if ( rv )
       return rv;

  rv = ActiveSession();

  if ( rv )
       return rv;

  rv = SetSessionPriv();

  if ( rv )
       return rv;

  m_is_active = 1;

  IpmiLog( "RMCP session is up.\n" );

  return 0;
}


// Send the final close session to shut the connection down.

void
cIpmiConLan::SendCloseSession()
{
  cIpmiMsg   msg( eIpmiNetfnApp, eIpmiCmdCloseSession );
  cIpmiAddr si( eIpmiAddrTypeSystemInterface );

  IpmiSetUint32( msg.m_data, m_session_id );
  msg.m_data_len = 4;

  Send( si, msg );
}


void
cIpmiConLan::IfClose()
{
  if ( m_fd >= 0 )
     {
       SendCloseSession();
       m_inuse = 0;
       m_is_active = 0;

       close( m_fd );
       m_fd = -1;

       if ( m_auth_method )
          {
            delete m_auth_method;
            m_auth_method = 0;
          }
     }
}


void
cIpmiConLan::Reconnect()
{
  IpmiLog( "lan reconnection in progress.\n" );

  if ( m_fd >= 0 )
     {
       m_is_active = 0;

       close( m_fd );
       m_fd = -1;
     }

  do
       sleep( 1 );
  while( IfOpen() );

  IpmiLog( "lan reconnection done.\n" );
}


void
cIpmiConLan::IfEvent()
{
  cIpmiAddr     addr;
  cIpmiMsg      msg;
  tResponseType res;

  assert( m_inuse == 0 );

  do
       res = WaitForResponse( 0, addr, msg );
  while( res & eResponseEvent );
}
