/*
 * ipmi_sel.cpp
 *
 * Copyright (c) 2003 by FORCE Computers
 *
 * Note that this file is based on parts of OpenIPMI
 * written by Corey Minyard <minyard@mvista.com>
 * of MontaVista Software. Corey's code was helpful
 * and many thanks go to him. He gave the permission
 * to use this code in OpenHPI under BSD license.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  This
 * file and program are licensed under a BSD style license.  See
 * the Copying file included with the OpenHPI distribution for
 * full licensing terms.
 *
 * Authors:
 *     Thomas Kanngieser <thomas.kanngieser@fci.com>
 */

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "ipmi_cmd.h"
#include "ipmi_sel.h"
#include "ipmi_mc.h"
#include "ipmi_log.h"
#include "ipmi_utils.h"
#include "ipmi_event.h"


cIpmiSel::cIpmiSel( cIpmiMc *mc, unsigned int lun )
  : m_mc( mc ), m_lun( lun ),
    m_major_version( 0 ), m_minor_version( 0 ),
    m_entries( 0 ),
    m_last_addition_timestamp( 0 ),
    m_last_erase_timestamp( 0 ),
    m_overflow( false ),
    m_supports_delete_sel( false ), m_supports_partial_add_sel( false ),
    m_supports_reserve_sel( false ), m_supports_get_sel_allocation( false ),
    m_fetched( false ), m_reservation( 0 ),
    m_sels_changed( false ), m_events( 0 ), 
    m_num_sels( 0 ), m_del_sels( 0 )
{
  assert( lun < 4 );
}


cIpmiSel::~cIpmiSel()
{
  FreeEvents();
}


void
cIpmiSel::FreeEvents()
{
  while( m_events )
     {
       cIpmiSelEventHolder *h = (cIpmiSelEventHolder *)m_events->data;
       m_events = g_list_remove( m_events, h );

       delete h;
     }
}


void
cIpmiSel::FreeDeletedEvents()
{
  GList *list = m_events;

  while( list )
     {
       cIpmiSelEventHolder *h = (cIpmiSelEventHolder *)list->data;
       GList *next = g_list_next( list );

       m_events = g_list_remove( m_events, h );

       delete h;

       list = next;
     }
}


int
cIpmiSel::SelClear( cIpmiMc *mc )
{
  cIpmiMsg cmd_msg;
  cIpmiMsg rsp;
  int rv;

  // do a reservation only when needed
  if (    m_supports_reserve_sel 
       && m_reservation == 0 )
     {
       rv = Reserve();

       if ( rv )
            return rv;
     }

  IpmiLog( "clear SEL.\n" );

  cmd_msg.m_netfn    = eIpmiNetfnStorage;
  cmd_msg.m_cmd      = eIpmiCmdClearSel;
  cmd_msg.m_data_len = 6;
  IpmiSetUint16( cmd_msg.m_data, m_reservation );
  cmd_msg.m_data[2]  = 'C';
  cmd_msg.m_data[3]  = 'L';
  cmd_msg.m_data[4]  = 'R';
  cmd_msg.m_data[5]  = 0xaa;

  rv = mc->SendCommand( cmd_msg, rsp, m_lun );

  if ( rv )
       return rv;

  if ( rsp.m_data[0] == 0 )
     {
       // Success!  We can free the data.
       FreeDeletedEvents();
       m_del_sels = 0;
     }

  return 0;
}


int
cIpmiSel::GetInfo( unsigned int &fetched_num_sels )
{
  cIpmiMsg cmd_msg( eIpmiNetfnStorage, eIpmiCmdGetSelInfo );
  cIpmiMsg rsp;
  int rv;
  unsigned int add_timestamp;
  unsigned int erase_timestamp;

  // Fetch the repository info.
  rv = m_mc->SendCommand( cmd_msg, rsp, m_lun );

  if ( rv )
     {
       IpmiLog( "could not send get sel info: 0x%02x !\n", rv );
       return rv;
     }

  if ( rsp.m_data[0] != 0 )
     {
       IpmiLog( "IpmiSelGetInfo: IPMI error from SEL info fetch: %x !\n",
                rsp.m_data[0]);

       return EINVAL;
     }

  if ( rsp.m_data_len < 15 )
     {
       IpmiLog( "handle_sel_info: SEL info too short !\n");
       return EINVAL;
     }

  // Pull pertinant info from the response.
  m_major_version               = rsp.m_data[1] & 0xf;
  m_major_version               = (rsp.m_data[1] >> 4) & 0xf;
  fetched_num_sels              = IpmiGetUint16( rsp.m_data + 2 );
  m_overflow                    = (rsp.m_data[14] & 0x80) == 0x80;
  m_supports_delete_sel         = (rsp.m_data[14] & 0x08) == 0x08;
  m_supports_partial_add_sel    = (rsp.m_data[14] & 0x04) == 0x04;
  m_supports_reserve_sel        = (rsp.m_data[14] & 0x02) == 0x02;
  m_supports_get_sel_allocation = (rsp.m_data[14] & 0x01) == 0x01;

  add_timestamp   = IpmiGetUint32( rsp.m_data + 6 );
  erase_timestamp = IpmiGetUint32( rsp.m_data + 10 );

  // If the timestamps still match, no need to re-fetch the repository
  if (    m_fetched
       && (add_timestamp   == m_last_addition_timestamp )
       && (erase_timestamp == m_last_erase_timestamp    ) )
     {
       // If the operation completed successfully and everything in
       // our SEL is deleted, then clear it with our old
       // reservation.
       if ( (m_num_sels == 0) && m_events )
          {
	    // We don't care if this fails, because it will just
	    // happen again later if it does.
	    rv = SelClear( m_mc );

	    if ( rv )
                 return rv;
          }

       // no need to read sel
       return -1;
    }

  m_last_addition_timestamp = add_timestamp;
  m_last_erase_timestamp    = erase_timestamp;

  m_sels_changed = true;
  m_fetched = true;

  return 0;
}


int
cIpmiSel::Reserve()
{
  cIpmiMsg cmd_msg( eIpmiNetfnStorage, eIpmiCmdReserveSel );
  cIpmiMsg rsp;
  int rv;

  // Get a reservation.
  rv = m_mc->SendCommand( cmd_msg, rsp, m_lun );

  if ( rv )
     {
       IpmiLog( "cannot send reserve sel: 0x%02x !\n", rv );
       return rv;
     }

  if ( rsp.m_data[0] != 0 )
     {
       IpmiLog( "sel_handle_reservation: Failed getting reservation !\n" );
       return ENOSYS;
     }

  if ( rsp.m_data_len < 3 )
     {
       IpmiLog( "sel_handle_reservation: got invalid reservation length !\n" );
       return EINVAL;
     }

  m_reservation = IpmiGetUint16( rsp.m_data + 1 );

  return 0;
}


cIpmiSel::cIpmiSelEventHolder *
cIpmiSel::FindEvent( unsigned int recid )
{
  GList *list = m_events;

  while( list )
     {
       cIpmiSelEventHolder *h = (cIpmiSelEventHolder *)list->data;
       
       if ( h->m_event.m_record_id == recid )
            return h;

       list = g_list_next( list );
     }

  return 0;
}


GList *
cIpmiSel::ClearSel( GList *new_events )
{
  if ( m_num_sels && m_events )
     {
       /* We don't care if this fails, because it will just
          happen again later if it does. */
       SelClear( m_mc );
     }

  return new_events;
}


GList *
cIpmiSel::GetEvents()
{
  int fetch_retry_count = 0;
  int next_rec_id;
  int rv = 0;
  GList *new_events = 0;

  IpmiLog( "reading SEL.\n" );

  while( 1 )
     {
       unsigned int num;
       
       fetch_retry_count++;

       m_reservation = 0;

       if ( fetch_retry_count >= dMaxSelFetchRetries )
          {
            IpmiLog( "too many lost reservations in SEL fetch !\n");
            return ClearSel( new_events );
          }

       rv = GetInfo( num );

       if ( rv || num == 0 )
            return ClearSel( new_events );

       if ( m_supports_reserve_sel )
          {
            rv = Reserve();

            if ( rv )
                 continue;
          }

       next_rec_id = 0;

       while( 1 )
          {
            bool                 event_is_new = false;
            cIpmiSelEventHolder *holder;
            cIpmiEvent           del_event;
            unsigned int         record_id;

            // read record
            cIpmiMsg cmd_msg( eIpmiNetfnStorage, eIpmiCmdGetSelEntry );
            cIpmiMsg rsp;

            cmd_msg.m_data_len = 6;
            IpmiSetUint16( cmd_msg.m_data, m_reservation);
            IpmiSetUint16( cmd_msg.m_data+2, next_rec_id );
            cmd_msg.m_data[4] = 0;
            cmd_msg.m_data[5] = 0xff;

            rv = m_mc->SendCommand( cmd_msg, rsp, m_lun );

            if ( rv )
               {
                 IpmiLog( "Could not send SEL fetch command: %x !\n", rv );
                 return ClearSel( new_events );
               }

            if ( rsp.m_data[0] == eIpmiCcInvalidReservation )
               {
                 // We lost our reservation, restart the operation.  Only do
                 // this so many times, in order to guarantee that this
                 // completes.
                 IpmiLog( "SEL reservation lost %d !\n", fetch_retry_count );
                 m_reservation = 0;

                 break;
               }

            if ( rsp.m_data[0] != 0 )
               {
                 IpmiLog( "IPMI error from SEL fetch: %x !\n",
                          rsp.m_data[0] );

                 return ClearSel( new_events );
               }

            next_rec_id = IpmiGetUint16( rsp.m_data + 1 );

            del_event.m_mc        = m_mc;
            del_event.m_record_id = IpmiGetUint16( rsp.m_data + 3 );
            del_event.m_type      = rsp.m_data[5];
            memcpy(del_event.m_data, rsp.m_data + 6, 13 );

            IpmiLog( "read event: " );
            del_event.Log();

            record_id = IpmiGetUint16( rsp.m_data+3 );
            holder = FindEvent( record_id );

            if ( !holder )
               {
                 holder = new cIpmiSelEventHolder;

                 if ( !holder )
                    {
                      IpmiLog( "Could not allocate log information for SEL" );
                      return ClearSel( new_events );
                    }

                 m_events = g_list_append( m_events, holder );

                 holder->m_event = del_event;
                 holder->m_deleted = false;
                 event_is_new = true;
                 m_num_sels++;
               }
            else if ( del_event.Cmp( holder->m_event ) != 0 )
               {
                 // It's a new event in an old slot, so overwrite the old
                 // event.
                 holder->m_event = del_event;
                 if ( holder->m_deleted )
                    {
                      holder->m_deleted = false;
                      m_num_sels++;
                      m_del_sels--;
                    }

                 event_is_new = true;
               }

            if ( event_is_new )
               {
                 cIpmiEvent *e = new cIpmiEvent;
                 *e = del_event;
                 new_events = g_list_append( new_events, e );
               }
            else
                 IpmiLog( "SEL record already processed.\n" );

            if ( next_rec_id == 0xFFFF )
               {
                 // If the operation completed successfully and everything in
                 // our SEL is deleted, then clear it with our old
                 // reservation.
                 return ClearSel( new_events );
               }
          }
     }

  // not reached
  return 0;
}


int
cIpmiSel::AddEvent( cIpmiEvent *new_event )
{
  cIpmiSelEventHolder *holder;

  holder = FindEvent( new_event->m_record_id );

  if ( !holder )
     {
       holder = new cIpmiSelEventHolder;

       if ( !holder )
            return ENOMEM;

       m_events = g_list_append( m_events, holder );

       holder->m_event = *new_event;
       holder->m_deleted = false;
       m_num_sels++;
     }
  else
     {
       holder->m_event = *new_event;

       if ( holder->m_deleted )
          {
	    holder->m_deleted = false;
	    m_num_sels++;
	    m_del_sels--;
          }
     }

  return 0;
}
